/*
 * IBM Confidential 
 * OCO Source Materials 
 * 5724-U98, 5655-V48, 5724-U50, 5655-V46  
 * The source code for this program is not published or other- 
 * wise divested of its trade secrets, irrespective of what has 
 * been deposited with the U. S. Copyright Office.
 */
header {
	package com.ibm.nex.parser.oim.distributed;
}

/*
 * OEF Distributed Parser
 */
{ @SuppressWarnings("all") }
class DistributedParser extends Parser;

options {
	buildAST = true;
	k = 4;
	defaultErrorHandler=false;
}

tokens {
	// These tokens are defined because ANTLR 2.76 does not automatically generate constants for literals that contain a number. 
	LITERAL_Q1="Q1";
	LITERAL_Q2="Q2";
	LITERAL_DB2="DB2";
	LITERAL_VALUE1="VALUE1";
	LITERAL_VALUE2="VALUE2";
	LITERAL_DB2CS="DB2CS";
	LITERAL_DB2MF="DB2MF";
	LITERAL_TYPE1="TYPE1";		
	LITERAL_TYPE2="TYPE2";		
	LITERAL_TYPE3="TYPE3";		
	LITERAL_TYPE4="TYPE4";
	LITERAL_WEEKEND1="WEEKEND1";
	LITERAL_WEEKEND2="WEEKEND2";
}

{		
	/**
	 * The characters that are invalid for an optim name.
	 * NOTE: This string comes from Optim Distributed code - rtrdba.cpp, gInvalidChars variable.
	 * In addition, the '%' character is invalid for names on import and export.
	 * Whitespace is not explicitly listed because tokens have been definied for whitespace, and such tokens
	 * should be skipped automatically.
	 */
	protected static final String INVALID_OPTIM_NAME_CHARACTERS = "%~`!^&*()+=|{}[]:;\"\'<>,.?/\\";	

	/**
	 * Determines if the token is a valid optim identifier.
	 */
	protected boolean isValidOptimIndentifierToken(Token token){
		boolean isValidToken = true;		
		String tokenText = token.getText();
		// Valid values are letters, '@', '#', '$', and '_'.  Digits are valid except as the first character.
		for(int charIndex=0; charIndex<tokenText.length(); charIndex++){
			char tokenCharacter = tokenText.charAt(charIndex);
			if (!Character.isLetter(tokenCharacter)){
				if ((charIndex > 0) && !Character.isDigit(tokenCharacter)){
					switch(tokenCharacter){
						case '@':
						case '#':
						case '$':
						case '_':
							break;
						default:
							return false;
					}
				} 
			}
		}
		return true;
	}

	/**
	 * Determines if the token is a valid optim name.
	 */
	protected boolean isValidUnsignedInteger(Token token){
		return isValidUnsignedInteger(token.getText());
	}
	
	protected boolean isValidUnsignedInteger(String tokenText){
		if (tokenText.length() == 0){
			return false;
		}
		for(int charIndex=0; charIndex<tokenText.length(); charIndex++){
			if (!Character.isDigit(tokenText.charAt(charIndex))){
				return false;
			}
		}
		return true;		
	}

	/**
	 * Determines if the token is a valid optim name.
	 */
	protected boolean isValidOptimNameToken(Token token){
		String tokenText = token.getText();
		for(int charIndex=0; charIndex<tokenText.length(); charIndex++){
			char tokenCharacter = tokenText.charAt(charIndex);
			// If the character is invalid for an optim name, return false.
			if (INVALID_OPTIM_NAME_CHARACTERS.indexOf(tokenCharacter) != -1){
				return false;
			}
		}		
		return true;		
	}
	
	/**
	 * Determines if the token is a valid variable delimiter.
	 */
	protected boolean isValidVariableDelimiter(Token token, boolean allowZero){
		boolean isValidToken = false;
		String tokenText = token.getText();
		if (tokenText.length() == 1){
			char tokenCharacter = tokenText.charAt(0);
			switch(tokenCharacter){
				case '@':
				case '#':
				case '$':
					isValidToken = true;
					break;
				case '0':
					isValidToken = allowZero;
					break;
			}
		}
		return isValidToken; 
	}
	
	/**
	 * Determines if the token is a valid hex string.
	 */
	protected boolean isValidHexString(String string){
		if (string.isEmpty()){
			return false;
		}
		string = string.toLowerCase();
		final String HEX_CHARS = "abcdef";
		for(int charIndex=0; charIndex<string.length(); charIndex++){
			if (!Character.isDigit(string.charAt(charIndex)) && !HEX_CHARS.contains(string.substring(charIndex, charIndex+1))){
				return false;
			}
		}
		return true;		
	}
	
	/**
	 * Handles cases where a keyword may be in place of a generic optim token.
	 * possibleTokenTypes: Other possible token types.  GENERICOPTIMTOKEN is assumed to be a valid type.
	 * throws: TokenStreamException if LA fails, RecognitionException if the token really is invalid. 
	  
	   Additional details:

	   This method handles cases where a keyword is really an GENERICOPTIMTOKEN.
			
	   Most grammars to not allow a name to be a keyword.  For example, in Java, you cannot have an int named "int".
	   However, Optim Export Files *do* allow a name to be a keyword.  For example, an export file can have a 
	   statement to create an Access Definition that begins with "CREATE AD AD.AD" where "AD" is a keyword, 
	   but "AD.AD" is the name of the Access Definition.
	   
	   To handle this condition, this method first checks to see if the token is one of the expected types.
	   GENERICOPTIMTOKEN is always included in the list, but in some cases other tokens may need to be checked.
	   The explanation for when you would need other tokens will follow in a moment.
	   
	   If the token is not an expected type, then we check to see if the token is a keyword 
	   *that has the properties* of an GENERICOPTIMTOKEN.  If so, we force the token to behave like 
	   a GENERICOPTIMTOKEN.
	   
	   All keywords in an OEF are simple strings of letters (and sometimes an underscore).  
	   Therefore, to determine if the keyword is really an identifier, the code just checks
	   that the keyword starts with a letter.
	   
	   This fix has a potential side-effect.  If the creator of an OEF forgets a name (for example, 
	   something that must follow a keyword), then the parser will see the next intended keyword, 
	   force it to be a different type, and accept it.
	   
	   For example, this piece of an Access Definition:
	   
	   		START CUSTOMERS ADDTBLS N 
	   
	   Includes keyword START for the start table, whose name is CUSTOMERS.  
	   It also includes keyword ADDTBLS with a value of N.
	   
	   If the user modified this in error as follows:
	   
	   		START ADDTBLS N
	   
	   Then the parser would read the start table's name as ADDTBLS, and fail on the "N" following ADDTBLS, claiming
	   that it could not find the expected token ADDTBLS.  
	   
	   In some cases, we need to call this method with a list of other expected keyword types.  
	   The reason for this is to handle ANTLR's parser code generation.
	   
	   In cases where alternatives are presented (e.g., "rule1 | rule2"), ANTLR generates code that looks ahead 
	   to determine which rule to follow.  The lookahead is based on tokens, and therefore the rule code is not
	   executed before the token is analyzed.
	   
	   To work around this, we simply add a call to this method and list all of the expected tokens.  
	   See variableDefaultValue for an example.  NOTE: It is not necessary to include GENERICOPTIMTOKEN in the list.
	   That token is assumed. 
	   
	   NOTE: I've tested with making the token in this rule match any token using the "." operator.  
	   The problem there is that too many tokens are allowed.  In a rule "(genericOptimToken)+ RIGHTPAREN", 
	   the RIGHTPAREN is treated as a genericOptimToken, and then fails the validation.
			
	   Another possible solution is to use a rule that lists *all* literal tokens.  Then, this rule would be 
	   "GENERICOPTIMTOKEN | anyLiteral".  However, this would be a maintenance nightmare.
			
	*/							 
	protected void handlePossibleGenericOptimToken(int [] possibleTokenTypes) throws TokenStreamException, RecognitionException {
		handlePossibleGenericOptimToken(1, possibleTokenTypes);
	}
	
	protected void handlePossibleGenericOptimToken(int tokenPosition, int [] possibleTokenTypes) throws TokenStreamException, RecognitionException {
		int nextTokenType = LA(tokenPosition);
		// If the token is already one of the possible token types, then do nothing else.
		if (possibleTokenTypes != null){
			for(int tokenType : possibleTokenTypes){
				if (nextTokenType == tokenType){
					return;
				}
			}
		}
		// If the token is not a GENERICOPTIMTOKEN, then check to see if it is a keyword that could be changed into a GENERICOPTIMTOKEN. 
		if (nextTokenType != GENERICOPTIMTOKEN) {
			Token token = LT(tokenPosition);
			if (isValidOptimNameToken(token)) {
				token.setType(GENERICOPTIMTOKEN);
			} else {
				throw new RecognitionException("Invalid value for genericOptimToken rule.", getFilename(), token.getLine(), token.getColumn());				
			}
		}	
	}
	
	/**
	 * Handles look-ahead for a multi-part name with variable number of name parts.
	 * This method is invoked for both identifiers and names, but since name is the
	 * least restrictive, we'll check for name validity and let the actual rule which
	 * has called this method validate each part.
	 */
	protected void handleMultiPartNameLookAhead() throws TokenStreamException, RecognitionException {
		int lookAheadTokenType;
    	handlePossibleGenericOptimToken(1, new int[] {QUOTEDSTRING});
        if (LA(2) == DOT) {
	    	handlePossibleGenericOptimToken(3, new int[] {QUOTEDSTRING});
        }
        if (LA(4) == DOT) {
	    	handlePossibleGenericOptimToken(5, new int[] {QUOTEDSTRING});
        }
	}
}

/*
 * CREATE
 */
create
	: ("CREATE"
		( accessDefinition
 		| archiveFileCollection
 		| archiveRequest
 		| calendar
		| currency
 		| columnMap
 		| columnMapProcedure
 		| convertRequest
 		| dbAlias
 		| deleteRequest
  		| extractRequest
        | insertRequest
        | loadRequest
  		| primaryKey
  		| relationship
  		| reportRequest  		
  		| restoreRequest  		
  		| storageProfile  		
  		| tableMap
  		)
	)+
	;

/*
 * AD
 */
accessDefinition
	: "AD" objectFullName
		accessDefinitionInternal
		SEMI
	;

accessDefinitionLocal
	: LEFTPAREN accessDefinitionInternal RIGHTPAREN
	;

accessDefinitionInternal
	: (description)?
		"SRCQUAL" databaseObjectQualifierOrQuotedString "START" qualifiedTableNameOrQuotedString
		"ADDTBLS" yesNoChoice "MODCRIT" yesNoChoice
		"ADCHGS" yesNoChoice "USENEW" yesNoChoice
		("USEFM" yesNoChoice)?
		pointAndShootState
		(group)?
		(variable)*
		(table)+
		(accessDefinitionRelationship)*
	    (archiveAction (COMMA archiveAction)*)?
		(defaultPaths)?	
	;

pointAndShootState
	: "PNSSTATE" noneLocalNamedChoice
		(localPointAndShootList)?
		("ROWLIST" DELIMITEDSTRING (optimServer)?)?
	;

group
	: "GRPCOL" columnNameOrQuotedString ("GRPROWS" unsignedInteger)? ("GRPVALS" unsignedInteger)?
	;

/**
 * An Access Definition variable
 */	
variable
	: "VAR" LEFTPAREN variableName
		"PRMPT" DELIMITEDSTRING
        (variableDefault)?
		RIGHTPAREN
	;

variableName
	: optimIdentifier
	;

/**
 * A variable default value.
 *
 * NOTE: A default recognized by Optim Distributed may have characters that are currently restricted by GENERICOPTIMTOKEN.  
 * A non-quoted default is valid as long 
 * as the string does not contain any of the following:
 *   a single quote
 * 	 a semicolon
 * 	 a left parenthesis
 * 	 a right parenthesis
 *   a comma
 *   a space character
 * However, rather than handle this unusual grammar situation,
 * for now the rule will only allow a GENERICOPTIMTOKEN or a quoted string.
 */
variableDefault
	: "DFLT" variableDefaultValue
	;

variableDefaultValue
	: {			
		handlePossibleGenericOptimToken(new int[] {QUOTEDSTRING});
	} (QUOTEDSTRING | genericOptimToken)
	; 

table
	: "TABLE" LEFTPAREN qualifiedTableNameOrQuotedString "ACCESS" tableAccessString "REF" yesNoChoice		
	  ("DAA" yesNoChoice)? 
	  ("UR" yesNoChoice)? 
	  ("EXTRFREQ" unsignedInteger)? 
	  ("EXTRLIMIT" unsignedInteger)?
	  predicateOperator 
	  "VARDELIM" variableDelimiter[false] "COLFLAG" yesNoChoice ("CORRELNAME" correlationID)?
	  (column)* 
	  (sqlClause)?
	  (sort)?
	  (archiveAction (COMMA archiveAction)*)?
	  (fileAttachment)*
	  (archiveIndex)*
	  ("EXTRROWID" accessByRowIdOption)?
	  RIGHTPAREN
	;

correlationID
	: optimName
	;

accessByRowIdOption
	: "N" | "D" | "A"
	;

/**
 * Describes table access.  Must contain one or more of the following characters:
 * 		S, U, I, D
 * and only in that order.
 */
tableAccessString 
	: "S" |
	  "U" |
	  "I" |
	  "D" |
	  "SU" |
	  "SI" |
	  "SD" |
	  "UI" |
	  "UD" |
	  "ID" |
	  "SUI" |
	  "SUD" |
	  "SID" |
	  "UID" |
	  "SUID"
	;

column
	: "COLUMN" LEFTPAREN columnNameOrQuotedString
		"DISP" yesNoChoice "ACCESS" columnAccessType "HEADING" headingType
		"NATIVELOB" yesNoChoice "EXTRACT" yesNoChoice
		("ASSOCIATION" columnAssociationName)? ("PRED" DELIMITEDSTRING)? RIGHTPAREN
	;

columnAssociationName
	: DOT fileExtension | columnName
	;

columnAccessType
	: "U" | "S"
	;

sort
	: "SORT" LEFTPAREN sortColumnName (COMMA sortColumnName)* RIGHTPAREN
	;

fileAttachment
	: "FILEATTACH" LEFTPAREN "PREFIX" columnNamePrefix
		("TRIGGER" columnName)?
		"STOP" yesNoChoice "DELFILE" yesNoChoice "INCL_DEFPATH" yesNoChoice
		"NAME_PARTS" LEFTPAREN fileAttachmentNameParts RIGHTPAREN
		("SEARCH_PATHS" LEFTPAREN (DELIMITEDSTRING)+ RIGHTPAREN)?
		RIGHTPAREN
	;

fileAttachmentNameParts {
		for (int i=1; LA(i) != RIGHTPAREN; i++){
			handlePossibleGenericOptimToken(i, new int[] {QUOTEDSTRING});
		}
	}
	: (columnNameOrQuotedString)+
	;

/**
 * A prefix for a column name - used in cases where a column name is constructed.
 * Essentially follows the same syntax as column, but we specify the rule for clarity and for the possibility of additional validation.
 */
columnNamePrefix
	: columnName
	;

archiveIndex
	: "ARCIDX" LEFTPAREN archiveIndexName (COMMA sortColumnName)+ RIGHTPAREN
	;

accessDefinitionRelationship
	: "REL" LEFTPAREN (DELIMITEDSTRING | relationshipName)
		"STATUS" statusType "USAGE" dormantInitialSelectedChoice
		"Q1" yesNoChoice "Q2" yesNoChoice
		"LIMIT" unsignedInteger "TYPE" relationshipType
		"PAR" qualifiedTableNameOrQuotedString ("PAR_ACCESS" defaultKeyScanChoice)? ("PAR_KEYLIMIT" unsignedInteger)?
		"CHI" qualifiedTableNameOrQuotedString ("CHI_ACCESS" defaultKeyScanChoice)? ("CHI_KEYLIMIT" unsignedInteger)?
		(accessDefinitionRelationshipIgnoreOptions)?
		RIGHTPAREN
	;

accessDefinitionRelationshipIgnoreOptions:
	"REL_IGNOREEMPTY"
	(
		( 
			LEFTPAREN 
			(accessDefinitionRelationshipIgnoreColumnSettingsVersion721)+ 
			"REL_CRITERIA" andOrChoice
			RIGHTPAREN
		)
	|
		(
			(accessDefinitionRelationshipIgnoreColumnSettings (PIPE accessDefinitionRelationshipIgnoreColumnSettings)*)? 
			"REL_CRITERIA" unsignedInteger
		)
	)
	;

accessDefinitionRelationshipIgnoreColumnSettingsVersion721
	: "COLUMN" LEFTPAREN 
	columnName
	"NULL" yesNoChoice
	"BLANK" yesNoChoice
	"ZEROLENGTH" yesNoChoice
	"NUMBER" ("N" | floatNumber)
	RIGHTPAREN
	;

accessDefinitionRelationshipIgnoreColumnSettings
	: columnName COLON unsignedInteger (COLON floatNumber)?
	;

relationshipType
	: "PST" |
      "PST_GENERIC" |
      "UNK" |
      "DB2" "UDB" |
      "DB2" "LUW" |
      "UDB" |
      "DB2" "z" SLASH "OS" |
      "Oracle" |
      "Sybase" |
      "ODBC" |
      "Informix" |
      "SQL" "Server" |
      "Database" |
      "Teradata" |
      "Netezza"
	;

/**
 * Used in Access Definitions and Table Maps.  Table Maps use Archive Actions only for Restore.
 */
archiveAction
	: "ARCHACTS" LEFTPAREN "ACTION" archiveActionType
	   ("SAMEAS" archiveActionType)?
	   ("DBALIAS" dbAliasName)?
	   (sqlClause)?
	   ("HOSTVAR" variableDelimiter[false])?
	   ("ADVAR" variableDelimiter[true])?
	   ("ON_ERROR" errorType)?
	   RIGHTPAREN
	;

defaultPaths
	: "DEFPATHS" LEFTPAREN (DELIMITEDSTRING)+ RIGHTPAREN
	;
	
/*
 * CM
 */
columnMap
	: "CM" objectFullName
		columnMapInternal
		SEMI
	;

columnMapLocal
	: LEFTPAREN columnMapInternal RIGHTPAREN
	;

/**
 * The common elements of the column map.
 * NOTE: The SRC and DEST keywords and values are required for named column maps,
 * but are not included in local column maps.
 * This syntactic difference will be handled in the load and save code.
 */
columnMapInternal
	: (description)?
	  (
		  (
			  validateRule
			  (optimServer)?
			  "DSTUSESSRC" trueFalseChoice
			  ("SRCEXTDSN" DELIMITEDSTRING)?
			  ("DSTEXTDSN" DELIMITEDSTRING)?
			  (
				"SRC" fullyQualifiedTableName 
				"DEST" fullyQualifiedTableName
			  )?
		  )
		  |
		  (
			  ("EXTDSN" DELIMITEDSTRING)?
		      "SRC" fullyQualifiedTableName 
		      "DEST" fullyQualifiedTableName
		      validateRule
		  )
	  )
	  LEFTPAREN columnExpressionAssignment (COMMA columnExpressionAssignment)* RIGHTPAREN	
	;

columnMapProcedure
	: "CMPROC" objectFullName
		columnMapProcedureInternal
		SEMI
	;

columnMapProcedureLocal
	: LEFTPAREN columnMapProcedureInternal RIGHTPAREN
	;

/**
 * The column map procedure.
 */
columnMapProcedureInternal
	: (description)?
	  "TEXT" DELIMITEDSTRING
	;

/**
 * A column expression assignment.
 * NOTE: This needs to be enhanced to handle column map procedures.
 */
columnExpressionAssignment {			
		// The first predicate is at position 1, the second is at position 3.
		handlePossibleGenericOptimToken(1, new int[] {QUOTEDSTRING, LITERAL_NOT_SPECIFIED});
		handlePossibleGenericOptimToken(3, new int[] {QUOTEDSTRING, LITERAL_NOT_SPECIFIED});
		// If this columnExpressionAssignment is followed by a comma, then look ahead and handle the next token, which may be a columnExpressionPredicate. This is necessary because the lookahead value k=4, and when the second predicate is evaluated, the first predicate of the next assignment is within 4 tokens away.
		if (LA(4) == COMMA){
			handlePossibleGenericOptimToken(5, new int[] {QUOTEDSTRING, LITERAL_NOT_SPECIFIED});			
		}		
	}	
    : ("NOT_SPECIFIED" | columnExpressionPredicate) EQUALS columnExpressionPredicate
      (
		"CMPROC" objectFullName
		| "PROC_LOCAL" columnMapProcedureLocal
	  )?
	;

/**
 * A column expression predicate.
 * name and floatNumber can conflict at k==2.  This could happen for a single integer without a plus sign.
 * For example, "3" could be either a name or a floatNumber.  
 * In such a case, name takes precedence because it comes first.
 */
columnExpressionPredicate
	: escapableQuotedString | columnName | floatNumber
	;

/**
 * One or more quoted strings next to each other.  
 * This accounts for either a quoted string, 
 * or a quoted string that can contain "escaped" single quotes 
 * - that is, single quotes that are doubled-up to indicate a single quote nested in a 
 * single-quoted string (dumb grammar, but that's how it exists).
 * To handle this case, we read multiple quoted strings.  For example,
 *
 * 		'TRANS EML(''ADDRESS, "email"'')'
 *
 * will be read as these tokens:
 *
 *		'TRANS EML('
 *		'ADDRESS, "email"'
 *		')'
 * 
 * The loader will have to merge them together. 
 */
escapableQuotedString
	: (QUOTEDSTRING)+
	;

/**
 * DB Alias
 */
dbAlias
	: "DBALIAS" dbAliasName
		(description)?
		"TYPE" typeChoice 
		"VER" unsignedInteger
		("IDCASE" caseChoice)?
		"CODEPAGE" unsignedInteger
		("DELIM" parenthesizedStringDelimiter)? 
		("DECSEP" LEFTPAREN decimalSeparator RIGHTPAREN)?
		"CONNECT" databaseServerName
		("DBQUAL" databaseQualifier)?		
		(	
			("ALITYPE" dbAliasObjectCreatorIDType ("ALICID" creatorID)?) |		
			("SYNTYPE" dbAliasObjectCreatorIDType ("SYNCID" creatorID)?)
		)?
		("TRITYPE" dbAliasObjectCreatorIDType ("TRICID" creatorID)?)?				
		("SPQUAL" storedProcedureQualifier)? 
		("NETSERVER" serverName)? 
		("DATABASE" databaseName)? 
		("USEDFLTDB" yesNoChoice)?
		("TABLESPACE" tableSpaceName)?
		"USEDFLTTBLSPC" yesNoChoice 
		("TABLEALLOC" unsignedInteger)?
		("INDEXALLOC" unsignedInteger)?
		"IDXTYPE" indexType
		("IDXCID" creatorID)?
		("IDXTBLSPCTYPE" indexTablespaceType)?
		("IDXTBLSPC" tableSpaceName)?
		"PASSWORDREQ" yesNoChoice
		"UNICODEDB" yesNoChoice
		SEMI
	;

/**
 * databaseServerName
 * 
 * This is either a connection string or an IP address.
 */
databaseServerName {			
		handlePossibleGenericOptimToken(null);
	}
	: databaseObjectName | ipAddress 
	;

ipAddress
	: unsignedInteger DOT unsignedInteger DOT unsignedInteger DOT unsignedInteger
	;

databaseQualifier
	: databaseObjectName
	;

storedProcedureQualifier
	: databaseObjectName
	;

tableSpaceName
	: databaseObjectName
	;

databaseName
	: databaseObjectName
	;

/*
 * PK
 */
primaryKey
	: "PK" primaryKeyName
		(description)?
		(baseCreatorID)?
		"COLS" LEFTPAREN columnName (COMMA columnName)* RIGHTPAREN
		SEMI
	;

/**
 * A primary key name.
 * This rule is similar to fullyQualifiedTableName, 
 * but the second part can be a STAR to account for generic primary keys.
 */
primaryKeyName
	: dbAliasName DOT (creatorID | STAR) DOT tableName
	;

/*
 * REL
 */
relationship
	: "REL" relationshipName
		(description)?
		"CHI" fullyQualifiedTableNameOrQuotedString
		"PAR" fullyQualifiedTableNameOrQuotedString
		(baseCreatorID)?
		LEFTPAREN relationshipPredicateAssignment (COMMA relationshipPredicateAssignment)* RIGHTPAREN
		SEMI
	;

relationshipPredicateAssignment {
		// Resolve any literal tokens that we really want to be strings.
		// The first predicate is at position 1, the second is at position 3.
		handlePossibleGenericOptimToken(1, new int[]{QUOTEDSTRING, DOT});
		handlePossibleGenericOptimToken(3, new int[]{QUOTEDSTRING, DOT});
	}
	: relationshipPredicate EQUALS relationshipPredicate
	;

/**
 * A relationship predicate.
 * name and floatNumber can conflict at k==2.  This could happen for a single integer without a plus sign.
 * For example, "3" could be either a name or a floatNumber.  
 * In such a case, name takes precedence because it comes first.
 */
relationshipPredicate {
		// Resolve any literal tokens that we really want to be strings.
		handlePossibleGenericOptimToken(new int[]{QUOTEDSTRING, DOT});
	}
	: "NULL" | 
	  "TRUE" | 
      "FALSE" | 
      QUOTEDSTRING | 
      columnName | 
      floatNumber
	;

/*
 * TM
 */
tableMap
	: "TM" objectFullName
		tableMapInternal
		SEMI	
	;
 
tableMapLocal
	: LEFTPAREN tableMapInternal RIGHTPAREN
	; 
 
tableMapInternal
	: (description)?
		"SRCQUAL" databaseObjectQualifierOrQuotedString 
	    "DESTQUAL" databaseObjectQualifierOrQuotedString
		("COLMAPID" columnMapID)?
		(
			("SRCAD" objectFullName | "SRCLOCALAD" | "SRCEXT" DELIMITEDSTRING)
			"SRCTYPE" mapSourceType		
			("DESTAD" objectFullName | "DESTLOCALAD" | "DESTDB" | "DESTEXT" DELIMITEDSTRING)
			"DESTTYPE" mapSourceType
		)?			
		validateRule
		("UNUSEDOBJ" yesNoChoice)?
		(optimServer)?
		LEFTPAREN tableAssignment (COMMA tableAssignment)* RIGHTPAREN
		("ASSEMBLIES" tableMapObjectAssignments)?
		("DEFAULTS" tableMapObjectAssignments)?
		("FUNCTIONS" tableMapObjectAssignments)?
		("PACKAGES" tableMapObjectAssignments)?
		("PARTITION_FUNCTIONS" tableMapObjectAssignments)?
		("PARTITION_SCHEMES" tableMapObjectAssignments)?
		("PROCEDURES" tableMapObjectAssignments)?
		("RULES" tableMapObjectAssignments)?
		("SEQUENCES" tableMapObjectAssignments)?
		("UDTYPES" tableMapObjectAssignments)?
		("VIEWS" tableMapObjectAssignments)?		
		(archiveAction (COMMA archiveAction)*)?
	;

columnMapID
	: objectID
	;

// NOTE: Column maps are excluded for now.  Any table map with a column map will fail syntax checking until column maps are implemented.
tableAssignment
	: tableAssignmentTerm EQUALS tableAssignmentTermOrNotSpecified
	  (ownedColumnMap)?
	  (LEFTPAREN archiveAction (COMMA archiveAction)* RIGHTPAREN)?
	;

tableAssignmentTerm
    : qualifiedTableNameOrQuotedString
    ;

tableAssignmentTermOrNotSpecified { 
		handleMultiPartNameLookAhead();
	} : ("NOT_SPECIFIED" | tableAssignmentTerm)
	;

tableMapObjectAssignments
	: LEFTPAREN tableMapObjectAssignment (COMMA tableMapObjectAssignment)* RIGHTPAREN
	;

tableMapObjectAssignment
	: objectAssignmentTerm EQUALS ("NOT_SPECIFIED" | objectAssignmentTerm)
	;

/**
 * An object assignment term.
 * NOTE: For simplicity, the syntax does not include the ability to determine
 * whether the maximum number of parts an object can have is 2 or 3.
 * See the comments in the qualifiedObjectName rule for more information.    
 */
objectAssignmentTerm
    : (QUOTEDSTRING | qualifiedObjectName)
    ;

/*
 * ARCH
 */
archiveRequest
	: "ARCH" objectFullName
		(description)?
		"AF" DELIMITEDSTRING
		("FAD" objectFullName)?
		("AFX" DELIMITEDSTRING)? 
		("GROUP" groupName)?
		("STORAGE_PROFILE" storageProfileName)?
		rowlimit
		databaseConnections
		"DELDBCONNECTIONS" unsignedInteger
		(optimServer)?
		"DEFER_DAA" yesNoChoice 
		"REVIEW_DELETE" yesNoChoice 
		deleteControlFileIfSuccessful		
		commonExtractOptions
		"CREATEREPORT" yesNoChoice
		ownedAccessDefinition
		(ownedReportRequest)?
		objectIncludes
		pointAndShootOverrideSettings
		("ROWLIST" DELIMITEDSTRING)?
		(localPointAndShootList)?
		(variables)?
		(
			controlFile
			discardRowLimit
			commitFrequency 
			lockTables
			includeLobs
			"COMPARE_BEFORE_DELETE" yesNoChoice
			"GENDELSTATISTIC" yesNoChoice
		)?
		("ALWAYSPROMPT" yesNoChoice)?
		(defaultObjectQualifier)?
		ignoreUnknown
		(objects)?
		("COLLECTIONS" LEFTPAREN objectFullName (COMMA objectFullName)* RIGHTPAREN)?
		(emailNotifySettings)?
		("DELETESTRATEGY" accessStrategyList)?
		SEMI
	;

/**
 * A locak report request.
 * NOTE: This is a stub for now - it will read everthing until the INCLPK keyword.
 * To save the request as a string in the model, traverse the tree until the INCLPK, then remove the leading and trailing parentheses.
 */
reportRequestLocal
	: (~("INCLPK"))*
	;

/*
 * Report request
 */
reportRequest
	: "REPT" (~SEMI)+ SEMI
	;

/*
 * Storage profile
 */
storageProfile
	: "STORPROF" (~SEMI)+ SEMI
	;

storageProfileName
	: objectName
	;

/*
 * Archive file collection
 */
archiveFileCollection
	: "AFC" (~SEMI)+ SEMI
	;

/*
 * Delete request
 */
deleteRequest
	: "DEL" objectFullName
	  (description)?
	  extractFile
	  controlFile
	  discardRowLimit
	  (commitFrequency)?
	  databaseConnections
	  lockTables
	  "REVIEWLIST" yesNoChoice
	  (optimServer)?
	  deleteControlFileIfSuccessful
	  "ARCHIVEFILE" yesNoChoice
	  includeLobs
	  "COMPARE_BEFORE_DELETE" yesNoChoice 
	  "VERIFYCOLSEXIST" yesNoChoice 
	  "GENERATE_STATISTICS" yesNoChoice
	  (emailNotifySettings)?
	  ("STRATEGY" accessStrategyList)?
	  ("DELBYROWID" yesNoChoice "DELROWIDTHREPER" unsignedInteger "DELROWIDTHREOPT" deleteRowIdInvalidThresholdAction)?
	  SEMI 
	;

includeLobs
	: "INCLUDE_LOBS" yesNoChoice
	;

deleteRowIdInvalidThresholdAction
	: "S" | "C" | "F"
	;

/*
 * CALENDAR
 */
calendar
    : "CALENDAR" objectName 
		(description)?
    	"MATCH" ("NEXT" | "PREVIOUS")
    	"WEEKEND1" calendarWeekendDay
    	"WEEKEND2" calendarWeekendDay
    	"DFLTSEP" LEFTPAREN . RIGHTPAREN
    	"DFLTYEAR" unsignedInteger
    	"MONTHS" LEFTPAREN
    	// Exactly 12 month names are specified.
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	monthNameData
    	RIGHTPAREN
    	(calendarDate)*
    	(calendarRule)*
		SEMI
	;

calendarWeekendDay {
		// Resolve any literal tokens that we really want to be integers, such as "2"
		handlePossibleGenericOptimToken(new int[]{LITERAL_None});
	}
	: ("None" | unsignedInteger)
	;

monthNameData
	: optimNameOrQuotedString COMMA optimNameOrQuotedString
	;

calendarDate
	: "DATE" LEFTPAREN
		calendarDateType
		(description)?
		// NOTE: "DAY" and "WHICH" could be written so that "WHICH" is included only if "DAY" is not "EASTER". However, in order to check for tokens that are actually integers (such as "2"), we have to make the rules separate. 
		"DAY" calendarDayType
		("WHICH" calendarWhichType)?
		"OFFSET" signedInteger
		"DURATION" unsignedInteger
		"ABSORB" yesNoChoice
		"RESOLVE" ("CLOSEST" | "PREVIOUS" | "NEXT" | "None")
		("MONTH" optimNameOrQuotedString)?
		("YEAR" unsignedInteger)?
		// Either "None" only, or a value followed by a "REOCCUR_VALUE".
		"REOCCUR_TYPE" ("None" | 
						("DAY" | "WEEK" | "MONTH" | "YEAR") "REOCCUR_VALUE" unsignedInteger
						)
		"REOCCUR_ENDDATE" ("None" | regionalFormatDate)
		RIGHTPAREN
	;

calendarDateType
	: optimName
	;

calendarDayType {
		// Resolve any literal tokens that we really want to be integers, such as "2"
		handlePossibleGenericOptimToken(new int[]{LITERAL_EASTER, LITERAL_WORKDAY, LITERAL_SPECIFIC});
	}
	: ("EASTER" | "WORKDAY" | "SPECIFIC" | unsignedInteger)
	;

calendarWhichType {
		// Resolve any literal tokens that we really want to be integers, such as "2"
		handlePossibleGenericOptimToken(new int[]{LITERAL_EVERY, LITERAL_LAST});
	}
	: ("EVERY" | "LAST" | unsignedInteger)
	;

calendarRule
	: "RULE" LEFTPAREN 
		objectName
		(description)?
		("AVOIDDATES" calendarDateTypeList)?		
		("SEARCHDATES" calendarDateTypeList)?
		"ADJUSTMENT" ("CLOSEST" | "PREVIOUS" | "NEXT")		
		RIGHTPAREN
	;

calendarDateTypeList
	: LEFTPAREN calendarDateType (COMMA calendarDateType)* RIGHTPAREN
	;

/*
 * CONV
 */
convertRequest
	: "CONV" objectFullName
		convertRequestInternal
		SEMI
	;

convertRequestLocal
	: LEFTPAREN convertRequestInternal SEMI
	;

/**
 * The common code for convert requests.
 * This is stubbed out for now.
 */
convertRequestInternal
	: (description)?
	  "SRCXF" DELIMITEDSTRING
	  ("DESTXF" DELIMITEDSTRING)?
	  controlFile
	  forceEditTableMap
	  "DELCNTLFILE" yesNoChoice
	  (discardRowLimit)?
	  ownedTableMap
	  showCurrency
	  showAge
	  (optimServer)?
	  (functionAging)?
	  (globalAging)?
	  (reportOptions)?
	  (currencyOptions)?
	  (emailNotifySettings)?
	  compressFile
	  "INCL_FILEATTACH" yesNoChoice
	  ( 
	  	("CSF" LEFTPAREN (csvTable)+ RIGHTPAREN) |
	    ("CSVF" LEFTPAREN csvFileOptions RIGHTPAREN) |
	    ("ECMF" LEFTPAREN ecmFileOptions RIGHTPAREN) |
		("XMLF" LEFTPAREN xmlFileOptions RIGHTPAREN) |
		("XMLBOF" LEFTPAREN (csvTable)+ RIGHTPAREN) |
		("HDFSF" LEFTPAREN hadoopFileOptions RIGHTPAREN)
	  )?
	  (convertActions)?
	;

csvTable
	: "TABLE" LEFTPAREN
	  fullyQualifiedTableName
	  ("CONSTRAINT" relationshipName)?
	  ("OWNER" fullyQualifiedTableName)?
	  ("OUTCOLS" LEFTPAREN columnName (COMMA columnName)* RIGHTPAREN)?
	  ("OUTPOSITION" LEFTPAREN unsignedInteger (COMMA unsignedInteger)* RIGHTPAREN)?
	  RIGHTPAREN
	;

commonConvertObjectOptions[boolean includeRowSupport]
	: "FILEMACRO" yesNoChoice
	  ("MACROVAL" LEFTPAREN (~RIGHTPAREN)+ RIGHTPAREN)?
	  commonConvertRowOptions[includeRowSupport]
	  "OBJPAIR" LEFTPAREN convertObjectData[includeRowSupport] (COMMA convertObjectData[includeRowSupport])* RIGHTPAREN
	;

commonConvertRowOptions[boolean includeRowSupport]
	: convertObjectDataRowOptions[includeRowSupport]
	;

convertObjectData[boolean includeRowSupport]
	: "OBJ" LEFTPAREN qualifiedTableName RIGHTPAREN
	  "PATH" LEFTPAREN DELIMITEDSTRING RIGHTPAREN
	  convertObjectDataRowOptions[includeRowSupport]
	;

// This method uses disambiguating semantic predicates to resolve ambiguity.
// From the ANTLR 2 manual:
//    Disambiguating predicates are always the first element in a production 
//    because they cannot be hoisted over actions, token, or rule references.
// Otherwise, the predicate would be a validating predicate, and an exception would be generated when it fails.
//
// A disambiguating semantic predicate clarifies which of multiple choices the parser should use. 
// If the value in the predicate is true, the first option is chosen. Otherwise, the second option is chosen.
// In this case, the second choice is empty, meaning that the choice is between the set of tokens in parentheses,
// or no extra tokens at all.
//
// Alternatively, we could make the rule convertObjectDataRowOptions optional by enclosing the tokens in (...)?
// However, this method allows us to enforce specific cases when the tokens are optional and when they are not.
convertObjectDataRowOptions[boolean includeRowSupport]
	: {includeRowSupport}? (
	  	"ROWCNT" LEFTPAREN unsignedInteger (COMMA unsignedInteger)* RIGHTPAREN
	  	"FILESIZE" LEFTPAREN unsignedInteger (COMMA unsignedInteger)* RIGHTPAREN
	  	"NUMFILES" LEFTPAREN unsignedInteger (COMMA unsignedInteger)* RIGHTPAREN
	  )
	  |
	;

commonConvertOutputOptions
	: "GENMDATA" yesNoChoice
	  (
	  	"BLABEL" LEFTPAREN (~RIGHTPAREN)+ RIGHTPAREN
	  	"ELABEL" LEFTPAREN (~RIGHTPAREN)+ RIGHTPAREN
	  	"HEADERDELIM" LEFTPAREN (~RIGHTPAREN)+ RIGHTPAREN
	  )?
	  "FIELDDELIM" LEFTPAREN DELIMITEDSTRING RIGHTPAREN
	  "STRDELIM" LEFTPAREN DELIMITEDSTRING RIGHTPAREN
	  "ESCCHAR" LEFTPAREN DELIMITEDSTRING RIGHTPAREN	
	;

commonConvertDataSettingsOptions
	: "LOBENABLED" yesNoChoice
	  ("LOBVAL" LEFTPAREN DELIMITEDSTRING RIGHTPAREN)?
	  "NULLCHECK" yesNoChoice
	  ("NULLVAL" LEFTPAREN (~RIGHTPAREN)+ RIGHTPAREN)?	
	;

ecmFileOptions
	: commonConvertObjectOptions[true]
	  "OPTIONS" LEFTPAREN
	  	commonConvertOutputOptions
	  	commonConvertDataSettingsOptions
	  	"ARCHFILENAME" yesNoChoice
	  	"ARCHDESC" yesNoChoice
	  	"ARCHDATE" yesNoChoice
	  	"ARCHGUID" yesNoChoice
	  RIGHTPAREN 
	;

csvFileOptions
	: commonConvertObjectOptions[true]
	  "OPTIONS" LEFTPAREN
	  	commonConvertOutputOptions
	  	commonConvertDataSettingsOptions
	  RIGHTPAREN 
	;

xmlFileOptions
	: commonConvertObjectOptions[true]
	  "OPTIONS" LEFTPAREN
	  	commonConvertDataSettingsOptions
	  RIGHTPAREN 
	;

hadoopFileOptions
	: commonConvertObjectOptions[false]
	  "OPTIONS" LEFTPAREN
	    commonConvertOutputOptions
	  	commonConvertDataSettingsOptions
	  RIGHTPAREN 
	;

convertActions
	: "CONVACTN" LEFTPAREN 
		"USEACTN" yesNoChoice
		(
			(convertAction)+
		)? 
		RIGHTPAREN
	;

convertAction 
	: "ACTION" LEFTPAREN
	  convertActionType
	  "USEPIPE" yesNoChoice
	  "ONERR" ("STOP" | "CONT")
	  "ACTNDELIM" variableDelimiter[false]
	  "UTILITY" DELIMITEDSTRING
	  "CMNDLNOPTS" DELIMITEDSTRING
	  (convertActionKeyValueData)*
	  RIGHTPAREN
	;

convertActionType
	: "SCP" |
	  "ECP" |
	  "STA" |
	  "ETA"
	;

convertActionKeyValueData
	: "KEYVALDATA" LEFTPAREN
	  "KEY" DELIMITEDSTRING
	  "VAL" DELIMITEDSTRING
	  "CLASS" convertActionKeyValueDataClassType
	  RIGHTPAREN
	;

convertActionKeyValueDataClassType
	: "USERDEF" |
	  "LOGFILELOC" |
	  "OPTIMVAR" |
	  "NAMEDPIPE"
	;

/*
 * CURRENCY
 */
currency
    : "CURRENCY" objectName 
		(description)?
		"DECPL" unsignedInteger
		"SHOW" ("EURO" | "ISO" | "USER")
		"ISOEDIT" yesNoChoice
		"PIVOT" unsignedInteger
		(currencyDateRange)+
		("TYPE1" currencyTypeCodeMap)?
		("TYPE2" currencyTypeCodeMap)?
		("TYPE3" currencyTypeCodeMap)?
		("TYPE4" currencyTypeCodeMap)?
     	SEMI
    ;

currencyDateRange
	:"RATE" LEFTPAREN
	regionalFormatDate regionalFormatDate
	(currencyRate)*
	RIGHTPAREN
	;

currencyRate
	: LEFTPAREN
	currencyType
	currencyDescription
	currencyType
	currencyDescription
	floatNumber
	RIGHTPAREN
	;

currencyDescription
	: optimNameOrQuotedString
	;

currencyTypeCodeMap
	: 	LEFTPAREN 
		(LEFTPAREN currencyDatabaseCode optimName RIGHTPAREN)+ 
		RIGHTPAREN
	;

currencyDatabaseCode {
		// Resolve any literal tokens that we really want to be integers, such as "2"
		handlePossibleGenericOptimToken(new int[]{QUOTEDSTRING, LITERAL_X, LITERAL_x});
		// Manually determine the difference between an unsigned integer and a hexadecimal literal to avoid nondeterminism.
		boolean isUnsignedInteger=isValidUnsignedInteger(LT(1));
	}
	:	{isUnsignedInteger}?unsignedInteger |
		hexadecimalLiteral |
		QUOTEDSTRING |
	;

/*
 * A hexadecimal literal has one of these formats:
 * 		X'1234567890ABCDEF'
 * or
 * 		0x1234567890ABCDEF  
 */
hexadecimalLiteral
	: 	hexadecimalLiteralFormat1 |
		hexadecimalLiteralFormat2
	;

hexadecimalLiteralFormat1{
		Token token = LT(2);
		if (LA(2) != QUOTEDSTRING){
			throw new RecognitionException("Invalid value for hexadecimalLiteralFormat2 rule.", getFilename(), token.getLine(), token.getColumn());			
		}
		String tokenText = token.getText();
		// Remove the quotes
		tokenText = tokenText.substring(1, tokenText.length()-1);
		if (!isValidHexString(tokenText)){
			throw new RecognitionException("Invalid value for hexadecimalLiteralFormat2 rule.", getFilename(), token.getLine(), token.getColumn());
		}
	}
	: ("X" | "x")QUOTEDSTRING
	;

hexadecimalLiteralFormat2  {
		Token token = LT(1);
		String tokenText = token.getText();
		tokenText = tokenText.toLowerCase();
		final String HEXADECIMAL_PREFIX = "0x";
		if (!tokenText.startsWith(HEXADECIMAL_PREFIX)){
			throw new RecognitionException("Invalid value for hexadecimalLiteralFormat2 rule.", getFilename(), token.getLine(), token.getColumn());
		}
		tokenText = tokenText.substring(HEXADECIMAL_PREFIX.length());
		if (tokenText.isEmpty()){
			throw new RecognitionException("Invalid value for hexadecimalLiteralFormat2 rule.", getFilename(), token.getLine(), token.getColumn());
		}
		if (!isValidHexString(tokenText)){
			throw new RecognitionException("Invalid value for hexadecimalLiteralFormat2 rule.", getFilename(), token.getLine(), token.getColumn());
		}
	}
	: genericOptimToken
	;


/*
 * EXTR
 */
extractRequest
	: "EXTR" objectFullName
		(description)?
		(optimServer)?
		extractFile
		ownedAccessDefinition
		pointAndShootOverrideSettings
		("PNSSTART" qualifiedTableNameWithQuotableParts)?
		("ROWLIST" fullPathName)?
		(localPointAndShootList)?
		(variables)?
		"OPTION" dataObjectsBothChoice
		objectIncludes
		ignoreUnknown
		commonExtractOptions
		rowlimit
		databaseConnections
		(defaultObjectQualifier)?
		(objects)?
		(ownedConvertRequest "CVDELXF" yesNoChoice)?
		(emailNotifySettings)?
		SEMI
	;

/*
 * UPIN
 */
insertRequest
	: "UPIN" objectFullName
		insertRequestInternal
		SEMI
	;

insertRequestLocal
	: LEFTPAREN insertRequestInternal RIGHTPAREN
	;

insertRequestInternal
	: 	(description)?
		(optimServer)?
		extractFile
		controlFile
		ownedTableMap
		discardRowLimit
		commitFrequency
		forceEditTableMap
		lockTables
		"UPDINS" insertProcessMethod
		"DELETEROWS"deleteBeforeInsertMethod		
		("DELCOMMIT" commitDeleteMethod)? 
		alwaysCallCreate
		"TRIGMODE" alwaysPromptNeverChoice 
		"CONSMODE" alwaysPromptNeverChoice
		showCurrency
		showAge
		deleteControlFileIfSuccessful
		processFileAttachments
		(functionAging)?
		(globalAging)?
		(reportOptions)?
		(currencyOptions)?
		(insertTableSettings)+
		(emailNotifySettings)?
		("FILE_ATTACH" LEFTPAREN directoryMap RIGHTPAREN)?
	;

forceEditTableMap
	: "FORCEEDITTM" yesNoChoice
	;

alwaysCallCreate
	: "ALWAYSCALLCREATE" yesNoChoice
	;	

/*
 * LOAD
 */
loadRequest
	: "LOAD" objectFullName
		loadRequestInternal
		SEMI
	;

loadRequestLocal
	: LEFTPAREN loadRequestInternal RIGHTPAREN
	;

loadRequestInternal
	: 	(description)?
		extractFile
		controlFile
		ownedTableMap
		forceEditTableMap
		"STOPONERROR" yesNoChoice
		"STOPONFIRSTCONVERTERROR" yesNoChoice 
		"MODE" processingMode
		alwaysCallCreate
		("ALWAYSESTIMATERESOURCES" yesNoChoice)?
		"DELCNTLFILE" yesNoChoice
		showCurrency
		showAge
		"PROCESS_FILEATTACH" ("F" | "C")
		(optimServer)?
		(functionAging)?
		(globalAging)?
		(reportOptions)?
		(currencyOptions)?
		loadDBAliases	
		(emailNotifySettings)?
	;

loadDBAliases
	: ( db2csLoadDBAlias
	  | db2mfLoadDBAlias
	  | informixLoadDBAlias
	  | oracleLoadDBAlias
	  | sqlServerLoadDBAlias
	  | sybaseLoadDBAlias
	  | teradataLoadDBAlias
	  | netezzaLoadDBAlias
	  )*
	;

db2csLoadDBAlias
	: "DB2CS" LEFTPAREN 
	  dbAliasName
	  "TYPE" ("I" | "R")
	  "LOAD" yesNoChoice
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?
	  ("USEPIPE" yesNoChoice)?
	  ("INLINELOBS" yesNoChoice)?
	  ("SCANLOBS" yesNoChoice)?
	  ("MLOADIMPDELETETASK" yesNoChoice)?
	  ("REMOTELOAD" yesNoChoice)?
	  ("NONRECVBLE" yesNoChoice)?	  	  
	  "EXCPFAIL" yesNoChoice
	  "EXCPCNST" yesNoChoice
	  ("EXCPCID" creatorID)?
	  ("LOCAL" DELIMITEDSTRING)?
	  ("REMOTE" DELIMITEDSTRING)?
	  ("TEMPPATH" DELIMITEDSTRING)?
	  "FILETYPE" ("X" | "A" | "E" | "I" | "S" | "L")
	  ("DELIMITER" db2csLoadDelimiter)?
	  ("ADDPARMS" DELIMITEDSTRING)?
	  "COPY" ("I" "IMAGE" fullDirectoryName | "A" "SESS" unsignedInteger | "N")
	  ("STATS" db2csLoadStatistics)?
	  ("COMMITFREQ" unsignedInteger)?
	  ("DISCARDLIMIT" unsignedInteger)?
	  ("STATSOPTS" unsignedInteger)?
	  ("TERANUMSESS" unsignedInteger)?
	  RIGHTPAREN
	;

// TODO: Modify to validate the delimiter.
// Simply accept any single *token* as a valid delimiter.
db2csLoadDelimiter
	: .
	;

db2mfLoadDBAlias
	: "DB2MF" LEFTPAREN 
	  dbAliasName
	  "TYPE" ("R" | "S")
	  ("LOCAL" DELIMITEDSTRING)?
	  "LOG" yesNoChoice
	  "RESETPEND" yesNoChoice
	  "ENFORCERI" yesNoChoice
	  "SINGLEDATA" yesNoChoice
	  "RUNSTATS" yesNoChoice
	  "RUNSTATSREPT" yesNoChoice
	  ("DISCARDLIMIT" unsignedInteger)?
	  ("CODEPAGE" unsignedInteger)?
	  "FTPFILES" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?	  
	  "FTPSUBMIT" yesNoChoice
	  "FTPREVIEW" yesNoChoice
	  "FTPSAVEJCL" yesNoChoice
	  "FTPPOL" yesNoChoice
	  ("FTPSERVER" ftpServerName)?	  
	  "FTPPORT" unsignedInteger
	  ("FTPUSERID" ftpUserId)?
	  ("FTPPASSWORD" ftpEncodedPassword)?
	  ("FTPQUALIFIER" dataSetQualifier)?
	  ("FTPTEMPLATE" DELIMITEDSTRING)?
	  RIGHTPAREN
	;

// This method uses disambiguating semantic predicates to resolve ambiguity.
// From the ANTLR 2 manual:
//    Disambiguating predicates are always the first element in a production 
//    because they cannot be hoisted over actions, token, or rule references.
// Otherwise, the predicate would be a validating predicate, and an exception would be generated when it fails.
//
// A disambiguating semantic predicate clarifies which of multiple choices the parser should use. 
// If the value in the predicate is true, the first option is chosen. Otherwise, the second option is chosen.
//
// Alternatively, we could make the rule convertObjectDataRowOptions optional by enclosing the tokens in (...)?
// However, this method allows us to enforce specific cases when the tokens are optional and when they are not.
ftpServerName {
		// Resolve the first token to be either an integer or part of a qualified domain name (in either case, a genericOptimToken).			
		handlePossibleGenericOptimToken(null);
		// Manually determine the difference between an unsigned integer and a string to avoid strings being misread as integers.
		boolean isUnsignedInteger=isValidUnsignedInteger(LT(1));		
	}
	:	{isUnsignedInteger}?(ipAddress) | 
		(qualifiedDomainName)
	;

qualifiedDomainName {			
		handlePossibleGenericOptimToken(null);
	}
	: serverName (DOT serverName)*
	;

ftpUserId
	: genericOptimToken
	;

ftpEncodedPassword
	: genericOptimToken
	;

dataSetQualifier
	: dataSetName
	;

dataSetName
	: genericOptimToken (DOT genericOptimToken)*
	;

db2csLoadStatistics
	: "S"
	| "O"
	| "I"
	| "D"
	| "SO"
	| "SI"
	| "SD"
	| "OI"
	| "OD"
	| "ID"
	| "SOI"
	| "SOD"
	| "SID"
	| "OID"
	| "SOID"
	;

informixLoadDBAlias
	: "INFORMIX" LEFTPAREN 
	  dbAliasName
	  "LOAD" yesNoChoice
	  "TYPE" ("I" | "R")
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?
	  ("USEPIPE" yesNoChoice)?
	  ("COMMITFREQ" unsignedInteger)?
	  ("DISCARDLIMIT" unsignedInteger)?
	  "STARTVIO" yesNoChoice
	  "DELETEVIO" yesNoChoice
	  ("LOCAL" (fullDirectoryName | directoryName))?
	  ("ADDPARMS" DELIMITEDSTRING)?
	  "TRIGMODE" alwaysPromptNeverChoice
	  "CONSMODE" alwaysPromptNeverChoice	  
	  RIGHTPAREN
	;

oracleLoadDBAlias
	: "ORACLE" LEFTPAREN
	  dbAliasName
	  "TYPE" ("I" | "R" | "A" | "T")
	  "LOAD" yesNoChoice
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?
	  ("USEPIPE" yesNoChoice)?	  
	  "SAVEDASD" yesNoChoice
	  "INLINELOBS" yesNoChoice
	  "DSCFILE" yesNoChoice
	  "USEDIRECTPATH" yesNoChoice
	  ("PARALLELLOADS" yesNoChoice)?
	  ("UNRECOVERABLE" yesNoChoice)?	  
	  "STARTVIO" yesNoChoice
	  ("DISCARDLIMIT" unsignedInteger)?
	  ("LOCAL" DELIMITEDSTRING)?
	  ("REMOTE" DELIMITEDSTRING)?
	  ("ADDPARMS" DELIMITEDSTRING)?
	  "TRIGMODE" alwaysPromptNeverChoice
	  "CONSMODE" alwaysPromptNeverChoice
	  "COMMITFREQ" unsignedInteger
	  ("DELIMITER" (QUOTEDSTRING | oracleLoadDelimiter))?
	  RIGHTPAREN
	;

// TODO: Modify to validate the delimiter.  Distributed does not validate, so it's not as urgent that SOA does either.
// Simply accept any single character as a valid delimiter.
oracleLoadDelimiter
	: .
	;

sybaseLoadDBAlias
	: "SYBASE" LEFTPAREN
	  dbAliasName
	  "LOAD" yesNoChoice
	  "TYPE" ("I" | "R")
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?
	  ("USEPIPE" yesNoChoice)?
	  "DSCFILE" yesNoChoice
	  ("DISCARDLIMIT" unsignedInteger)?		 
	  ("LOCAL" DELIMITEDSTRING)?
	  ("REMOTE" DELIMITEDSTRING)?
	  ("ADDPARMS" DELIMITEDSTRING)?
	  "TRIGMODE" alwaysPromptNeverChoice
	  "CONSMODE" alwaysPromptNeverChoice
	  ("PARTS" LEFTPAREN (sybasePartitionMapping)+ RIGHTPAREN)?	  	  
	  RIGHTPAREN
	;

sybasePartitionMapping
	: creatorID DOT tableName EQUALS sybasePartitionName
	;

sybasePartitionName
	: optimName
	;

sqlServerLoadDBAlias
	: "SQLSRV" LEFTPAREN
	  dbAliasName
	  "LOAD" yesNoChoice
	  "TYPE" ("I" | "R")
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?
	  ("USEPIPE" yesNoChoice)?	  
	  "DSCFILE" yesNoChoice
	  ("DISCARDLIMIT" unsignedInteger)?	   
	  ("LOCAL" DELIMITEDSTRING)?
	  ("REMOTE" DELIMITEDSTRING)?
	  ("ADDPARMS" DELIMITEDSTRING)?
	  "TRIGMODE" alwaysPromptNeverChoice
	  "CONSMODE" alwaysPromptNeverChoice
	  "WINAUTH" yesNoChoice
	  RIGHTPAREN
	;

teradataLoadDBAlias
	: "TERADATA" LEFTPAREN 
	  dbAliasName
	  "TYPE" ("I" | "R" | "D")
	  "UTIL" ("M" | "F")
	  "LOAD" yesNoChoice
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  "SOURCEEMPTY" yesNoChoice
	  "USEPIPE" yesNoChoice
	  "INLINELOBS" yesNoChoice
	  "SCANLOBS" yesNoChoice
	  "EXCPCREATE" yesNoChoice
	  ("LOCAL" DELIMITEDSTRING)?
	  "FILETYPE" ("S" | "L")
	  ("DELIMITER" teradataLoadDelimiter)?
	  ("ADDPARMS" DELIMITEDSTRING)?
	  ("CHECKPOINT" unsignedInteger)?
	  ("ERRORLIMIT" unsignedInteger)?
	  "TERANUMSESS" unsignedInteger
	  RIGHTPAREN
	;

// TODO: Modify to validate the delimiter.
// Simply accept any single *token* as a valid delimiter.
// NOTE: A double-quote is not a valid delimiter for Teradata, 
// and a single-quote is enclosed in double-quotes, making it a quoted string (also a single token). 
teradataLoadDelimiter
	: .
	;


netezzaLoadDBAlias
	: "NETEZZA" LEFTPAREN
	  dbAliasName
	  "LOAD" yesNoChoice
	  "TYPE" ("I" | "R")
	  "DELOK" yesNoChoice
	  "DELFAIL" yesNoChoice
	  ("SOURCEEMPTY" yesNoChoice)?
	  ("PRESCAN" yesNoChoice)?
	  ("USEPIPE" yesNoChoice)?
	  ("MAXERROR" unsignedInteger)?	   
	  ("LOCAL" DELIMITEDSTRING)?
	  "FILETYPE" ("A" | "E")
	  ("DELIMITER" netezzaLoadDelimiter)?	  
	  ("ADDPARMS" DELIMITEDSTRING)?
	  RIGHTPAREN
	;

// TODO: Modify to validate the delimiter.
// Simply accept any single *token* as a valid delimiter.
netezzaLoadDelimiter
	: .
	;


aging
	: 	"AGETYPE" ageType
		("YEARS" unsignedInteger)? 
		("MONTHS" unsignedInteger)?
		("WEEKS" unsignedInteger)? 
		("DAYS" unsignedInteger)? 
		("SPECIFICYEAR" unsignedInteger)?
		("SPECIFICDATE" standardFormatDate)?
		("TARGETSTART" regionalFormatDate)?
		("TARGETEND" regionalFormatDate)?
		("MULTIPLE" unsignedInteger)?
		("RULE" calendarRuleName)?
		("CALENDAR" calendarName)?
		("PIVOT" unsignedInteger)? 
		("INVALIDDATES" yesNoChoice)?
		("SKIPPEDDATES" yesNoChoice)?
	;

calendarName
	: objectName
	;

/**
 * A calendar rule name.
 * NOTE: Actual allowed values are not limited to optimName, but this will suffice for now.
 */
calendarRuleName
	: optimName
	;

currencyOptions
	: "CURRENCY_OPTION" LEFTPAREN
		"DEFAULT" currencyTableName
	  	("GLOBAL" currencyTableName
	  		("FROM" currencyType)?
			("TO" currencyType)?
			"TRIANG" trueFalseChoice
		)?
		RIGHTPAREN
	;

currencyTableName
	: QUOTEDSTRING | objectName
	;

/**
 * A currency type - used to describe a currency conversion from one type to another.
 * NOTE: Actual allowed values are not limited to optimName, but this will suffice for now.
 */
currencyType
	: optimName
	;

directoryMap
	: 	("DEFAULT" DELIMITEDSTRING)?
		("PATHMAP" LEFTPAREN DELIMITEDSTRING DELIMITEDSTRING RIGHTPAREN)*	
	;

emailNotifySettings
	: "EMAILNOTIFY" LEFTPAREN
	  ("SILENTMODEONLY")?
	  (emailNotifyEntry)*
	  RIGHTPAREN
	;

emailNotifyEntry
	: emailSendType emailAddress	
	;

emailSendType
	: "Always"
	| "Success"
	| "Failure"	
	;

emailAddress
	: QUOTEDSTRING | genericOptimToken (DOT genericOptimToken)+
	;

insertTableSettings
	: "TBL" LEFTPAREN 
		qualifiedTableNameOrQuotedString 
		"DELETEROWS" yesNoChoice 
		"UPDINS" insertTableMethod
		RIGHTPAREN
	;

reportOptions
	: "REPORT_OPTION" LEFTPAREN 
		"RPTERROR" trueFalseChoice 
		("MAXRUNERR" unsignedInteger)?
		("MAXTBLERR" unsignedInteger)?
		"RPTSUMMARY" trueFalseChoice
		"RPTINVALID" trueFalseChoice
		"RPTSKIPPED" trueFalseChoice
		RIGHTPAREN	
	;

/*
 * REST
 */
restoreRequest
	:   {
			boolean isRestoreActionInsert = false;
		}
        "REST" objectFullName
		(description)?
		"ACTION" isRestoreActionInsert=restoreActionType
		"SELECTION_MODE" requestSelectionMode
		"DELETE_XFFILES" yesNoChoice 
		"AUTO_GEN" yesNoChoice
		"CONTINUE_ON_ERROR" yesNoChoice
		compressFile
		(restoreFile)+
		(restoreProcessorEntry[isRestoreActionInsert])+
		("GLBLCRIT" LEFTPAREN selectionCriteria RIGHTPAREN)?
		(emailNotifySettings)?
		SEMI
	;

restoreFile
	: "FILE" LEFTPAREN
		DELIMITEDSTRING 
		("ARCHIVE_ID" unsignedInteger)?
		(optimServer)?
		(extractFile)?
		("AD_OVERRIDE" DELIMITEDSTRING)?
		pointAndShootState
		("ROWLIM" unsignedInteger)?
		"SELCRIT_USE" noneGlobalLocalChoice 
		(selectionCriteria)?
		RIGHTPAREN
	;

restoreProcessorEntry[boolean isRestoreActionInsert]
	: "PARM" LEFTPAREN
		(
			("LOCALDEF" localInsertOrLoad[isRestoreActionInsert]) 
		|   ("DEF" objectFullName)
		)
		("VALUE1" DELIMITEDSTRING)?
		("VALUE2" DELIMITEDSTRING)?
		RIGHTPAREN
	;

// This method uses semantic predicates to resolve ambiguity.
// From the ANTLR 2 manual:
//    Disambiguating predicates are always the first element in a production 
//    because they cannot be hoisted over actions, token, or rule references.
// Otherwise, the predicate would be a validating predicate, and an exception would be generated when it fails.  
localInsertOrLoad[boolean isRestoreActionInsert]
	: {isRestoreActionInsert}? insertRequestLocal 
	| loadRequestLocal 
	;

/*
 * Common
 */
accessStrategy
	: "TABLE" LEFTPAREN fullyQualifiedTableNameOrQuotedString accessStrategyType unsignedInteger RIGHTPAREN
	;

/**
 * Access strategy list.
 * NOTE: The list may be empty, with just parentheses.
 */
accessStrategyList
	: LEFTPAREN (accessStrategy)* RIGHTPAREN
	;

commitFrequency
	: "COMMITFREQ" unsignedInteger
	;

commonExtractOptions
	: compressFile
	  (
		"COMPRESSMODEL" unsignedInteger
		"COMPRESSACTIVE" yesNoChoice
		("DEFTHRESHOLD" unsignedInteger)?
		("TABLETHRESHOLD" LEFTPAREN tableThreshold (COMMA tableThreshold)* (COMMA)? RIGHTPAREN)?
	  )?
		"GENSTATISTIC" yesNoChoice
		processFileAttachments	
	;

tableThreshold
	: fullyQualifiedTableName unsignedInteger
	;

compressFile
	: "COMPRESSFILE" yesNoChoice
	;

controlFile
	: "CF" DELIMITEDSTRING
	;

databaseConnections
	: "DBCONNECTIONS" unsignedInteger
	;

defaultObjectQualifier
	: "OBJQUAL" databaseObjectQualifierOrQuotedString
	;

deleteControlFileIfSuccessful
	: "DELCF" yesNoChoice
	;

description
	: "DESC" DELIMITEDSTRING
	;

discardRowLimit
	: "DISCARDLIMIT" unsignedInteger
	;

extractFile
	: "XF" DELIMITEDSTRING
	;

functionAging
	: "FUNCTION_AGING" LEFTPAREN aging RIGHTPAREN
	;

globalAging
	: "GLOBAL_AGING" LEFTPAREN aging RIGHTPAREN
	;

ignoreUnknown
	: "IGNOREUNKNOWN" yesNoChoice
	;

localPointAndShootList
	: "LOCALRL" DELIMITEDSTRING
	;

lockTables
	: "LOCKTBLS" yesNoChoice
	;

objectIncludes
	: "INCLPK" yesNoChoice
	  "INCLFK" yesNoChoice
	  "INCLIDX" yesNoChoice
	  "INCLALIAS" yesNoChoice
	  "INCLASSEMBLY" yesNoChoice
	  "INCLFUNCTION" yesNoChoice
	  "INCLPACKAGE" yesNoChoice
	  "INCLPAR_FUNCTION" yesNoChoice
	  "INCLPAR_SCHEME" yesNoChoice		
	  "INCLPROCEDURE" yesNoChoice
	  "INCLSEQUENCE" yesNoChoice
	  "INCLTRIGGER" yesNoChoice
	  "INCLVIEW" yesNoChoice
	  ("INCLDEFAULT" | "INCLDEF") yesNoChoice
	  "INCLRULE" yesNoChoice
	  "INCLUDT" yesNoChoice
	  ("INCLMETHOD" yesNoChoice)?
	;

objects
	: "OBJECTS" LEFTPAREN objectAssignment (COMMA objectAssignment)* RIGHTPAREN
	;

objectAssignment
	: qualifiedObjectName EQUALS objectType
	;

ownedAccessDefinition
	: (("AD" objectFullName) | ("LOCALAD" accessDefinitionLocal))
	;

ownedColumnMap
	: (("CM" objectFullName) | ("LOCALCM" columnMapLocal))
	;

ownedConvertRequest
	: "CV" objectFullName 
	| "LOCALCV" convertRequestLocal
	;

ownedReportRequest
	: (("REPORT" objectFullName) | ("LOCALREPORT" reportRequestLocal))
	;

ownedTableMap
	: (("TM" objectFullName) | ("LOCALTM" tableMapLocal))
	;

processFileAttachments
	: "PROCESS_FILEATTACH" yesNoChoice
	;

pointAndShootOverrideSettings
	:"PNSOVERRIDE" yesNoChoice "PNSOPT" noneLocalNamedChoice
	;

predicateOperator
	: "PREDOP" andOrChoice
	;

rowlimit
	: "ROWLIMIT" unsignedInteger
	;

selectionCriteria
	: "TABLEOP" andOrChoice
		(selectionCriteriaTable)+
	;

selectionCriteriaTable
	: "TABLE" LEFTPAREN fullyQualifiedTableName
		predicateOperator
		(sqlClause)?
		(selectionCriteriaColumn)*
		RIGHTPAREN
	;

selectionCriteriaColumn
	: "COLUMN" LEFTPAREN columnName
		"OP" operator
		"PRED" DELIMITEDSTRING
		RIGHTPAREN
	;

showAge
	: "SHOWAGE" yesNoChoice
	;

showCurrency
	: "SHOWCURRENCY" yesNoChoice
	;

optimServer
	:"SERVERNAME" (BACKSLASH BACKSLASH)? optimServerName
	;

/**
 * A servername minus the allowed "\\" prefix.
 * NOTE: Actual allowed values are not limited to optimName, but this will suffice for now.
 */
optimServerName
	: serverName
	;

serverName {			
		handlePossibleGenericOptimToken(null);
	}
	: genericOptimToken
	;

sqlClause
	: "SQL" DELIMITEDSTRING
	;

validateRule
	: "VALRULES" moveCompareChoice
	;

variables
	: ("VARS" DELIMITEDSTRING)? "ALWAYSPROMPT" yesNoChoice
	;

baseCreatorID
	: "BASECREATORID" creatorID
	;

fullDirectoryName
	: windowsFullDirectoryName | unixFullDirectoryName
	;

fullPathName
	: windowsFullPathName | unixFullPathName
	;

directoryName
	: windowsDirectoryName | unixDirectoryName
	;

windowsFullDirectoryName
	: ( 
	  	(windowsDriveLetter COLON (BACKSLASH)?) 
	  	| (BACKSLASH BACKSLASH serverName BACKSLASH)
	  ) 
	  (windowsDirectoryName)?
	;

/**
 * A Windows full path name.
 */
windowsFullPathName
	:
	windowsFullDirectoryName windowsFilename 
	;

/**
 * A Windows drive letter.
 * NOTE: This is a horrible solution.
 *  
 * However, many of these tokens are already used elsewhere, and they therefore
 * supercede any given rule.  While a rule can have code in it to replace a token type,
 * there are cases where the token is analyzed before the rule can be processed.
 * The most frequent example is the choice operator "|".
 * In a rule with "ruleX | ruleY", the parser looks to see if the next token matches the
 * first token in ruleX or in ruleY.  If it's neither, ANTLR throws an exception - the code within
 * the rules never gets touched. 
 */
windowsDriveLetter
	: "A" | "B"	| "C" | "D"	| "E" 
	| "F" | "G" | "H" | "I" | "J" 
	| "K" | "L" | "M" | "N" | "O"
	| "P" | "Q"	| "R" | "S" | "T"
	| "U" | "V"	| "W" | "X"	| "Y"
	| "Z"	
	;

/**
 * A Windows directory.
 * This is a very simplified format.  There may be some files that will not be read correctly.
 */
windowsDirectoryName
	: (genericOptimToken BACKSLASH)* genericOptimToken (BACKSLASH)? 
	;

/**
 * A Windows filename.
 * This is a very simplified format.  There may be some files that will not be read correctly.
 */
windowsFilename
	: genericFilename
	;

unixFullDirectoryName
	: SLASH (unixDirectoryName)?
	;

/**
 * A Unix full path name.
 */
unixFullPathName
	: unixFullDirectoryName unixFilename
	;

/**
 * A Unix directory.
 */
unixDirectoryName
	: (genericOptimToken SLASH)* genericOptimToken (SLASH)?
	;

/**
 * A Unix filename.
 */
unixFilename
	: genericFilename
	;

genericFilename
	: genericOptimToken (DOT (genericOptimToken)?)?
	;

/**
 * A filename extension.
 * NOTE: Actual allowed values are not limited to optimName, but this will suffice for now.
 * This value can even conflict with the common token (see RtFileImport::FindToken in Optim Distributed).
 * In such a case, the Optim Distributed code should be modified to import/export quoted strings. 
 */
fileExtension
	: genericOptimToken
	;

fullyQualifiedTableNameOrQuotedString
	: {
    	handleMultiPartNameLookAhead();
    } (QUOTEDSTRING | fullyQualifiedTableName)
	;

/**
 * A fully qualified table name.
 * Similar to qualifiedTableName, except all three parts are required.
 */
fullyQualifiedTableName
	: dbAliasName DOT (QUOTEDSTRING | creatorID) DOT (QUOTEDSTRING | tableName)
	;


qualifiedTableNameOrQuotedString
	: {
    	handleMultiPartNameLookAhead();
    } (QUOTEDSTRING | qualifiedTableName)
	;

/**
 * A qualified table name.
 * Such a table name can consist of a DBAlias, creator ID, and table name, 
 * where the DBAlias and the creator ID are not required.
 */
qualifiedTableName
    : {
    	handleMultiPartNameLookAhead();
    } ((dbAliasName DOT)? creatorID DOT)? tableName
	;

/**
 * A qualified table name where the creator ID and/or table name may appear in quotes.
 */
qualifiedTableNameWithQuotableParts
    : {
    	handleMultiPartNameLookAhead();
    } ((dbAliasName DOT)? (QUOTEDSTRING | creatorID) DOT)? (QUOTEDSTRING | tableName)
	;

/**
 * A qualified object name.
 * Such an object name can consist of a DBAlias, creator ID, and database object name, 
 * where the DBAlias and the creator ID are not required.
 * 
 * NOTE: In some cases, DBAlias.ObjectName is a legal name.
 * This rule interprets such a name as CreatorID.ObjectName.
 * Since the valid values for creator ID are a superset of the valid values for a DBAlias,
 * this rule will suffice for now.   
 */
qualifiedObjectName
    : {
        handleMultiPartNameLookAhead();
    } ((dbAliasName DOT)? creatorID DOT)? databaseObjectName
    ;

/**
 * A database object qualifier or quoted string. 
 * Such an object name can consist of a DBAlias and creator ID, where the creator ID is not required.
 *
 * The method is necessary to allow tokens that can also be part of a name (like "AD").
 *
 * In this case, there are alternatives (e.g., "rule1 | rule2").  Rule code is not executed before the alternative is selected, so the rule code must come first.  
 * It cannot be in one of the alternatives.  See genericOptimToken for details.
 *
 * In order to make sure the rule code is executed first, the alternatives must be surrounded by parentheses.
 */
databaseObjectQualifierOrQuotedString
	: {
        handleMultiPartNameLookAhead();
    } (QUOTEDSTRING | databaseObjectQualifier)
	;

databaseObjectQualifier
	: dbAliasName (DOT creatorID)?
	;

dbAliasName
	: optimIdentifier
	;

creatorID
	: optimName
	;

/**
 * A database table name
 * NOTE: This does not account for tables that can have special characters in the names, such as spaces.
 * This situation will be addressed later, if needed.
 */
tableName
	: databaseObjectName
	;

/**
 * A database column name
 * NOTE: This does not account for columns that can have special characters in the names, such as spaces.
 * This situation will be addressed later, if needed.
 */
columnName
	: databaseObjectName
	;


columnNameOrQuotedString
	: {			
		handlePossibleGenericOptimToken(new int[] {QUOTEDSTRING});
	} (QUOTEDSTRING | columnName)
	;

optimNameOrQuotedString
	: {			
		handlePossibleGenericOptimToken(new int[] {QUOTEDSTRING});
	} (QUOTEDSTRING | optimName)
	;

/**
 * A database column name prepended by A (for ascending) or D (for descending).
 * NOTE: This does not account for columns that can have special characters in the names, such as spaces.
 * This situation will be addressed later, if needed.
 */
sortColumnName { 		
		Token token = LT(1);
		String tokenText = token.getText();
		boolean isValidToken = true;
		char firstCharacter = tokenText.charAt(0);
		switch(firstCharacter){
			case 'A':
			case 'D':
				break;
			default:
				isValidToken = false;
		}
		
		if (!isValidToken) {
			throw new RecognitionException("Invalid value for sortColumnName rule.", getFilename(), token.getLine(), token.getColumn());
		}		
	}
	: columnName
	;

/**
 * A relationship name.  This must satisfy the Optim Distributed method RtRDBA::ValidRelName
 */
relationshipName { 
		Token token = LT(1);
		String tokenText = token.getText();
		boolean isValidToken = true;
		char firstCharacter = tokenText.charAt(0);
		if (!Character.isLetter(firstCharacter)){
			switch(firstCharacter){
				case '@':
				case '#':
				case '$':
					break;
				default:
					isValidToken = false;
			}
		}
		
		if (!isValidToken) {
			throw new RecognitionException("Invalid value for relationshipName rule.", getFilename(), token.getLine(), token.getColumn());
		}	
	}
	: genericOptimToken
	;

/**
 * An archive index name.
 * The possible values are not actually restricted to the optimName rule. 
 */
archiveIndexName
	: optimName
	;
	
groupName
	: objectName
	;	

objectFullName
	: objectID DOT objectName
	;

objectID
	: optimIdentifier
	;

objectName
	: optimIdentifier
	;

/**
 * An identifier recognized by Optim.
 * This rule is intended to handle Optim Distributed methods RtRDBA::IsPSTNameValid 
 * and RtRDBA::ValidPSTIdentifier.
 * NOTE: It is recommended that simple rules refer to this rule as a way of clarifying how the identifier is being used.
 * For example, dbAliasName is an optimIdentifier.
 * Please do NOT use this rule directly from a larger definition, such as accessDefinition.
 * Doing so obscures the actual purpose of the rule.  
 */
optimIdentifier 
	: {
		Token token = LT(1);
		if (!isValidOptimIndentifierToken(token)){
			throw new RecognitionException("Invalid value for optimIdentifier rule.", getFilename(), token.getLine(), token.getColumn());			
		}
	} genericOptimToken
	;

optimName
	: genericOptimToken
	;

/**
 * A database object name.  Used for things that can exist in a database, like tables, columns, views, procedures, packages, etc.
 */
databaseObjectName
	: optimName
	;

/**
 * A name recognized by Optim.
 * This rule is intended to handle Optim Distributed method RtRDBA::IsNameValid.
 * NOTE: It is recommended that simple rules refer to this rule as a way of clarifying how the name is being used.
 * For example, creatorID is an optimName.
 * Please do NOT use this rule directly from a larger definition, such as accessDefinition.
 * Doing so obscures the actual purpose of the rule.  
 */
 genericOptimToken
	: {
		handlePossibleGenericOptimToken(null);
	} GENERICOPTIMTOKEN
	;

floatNumber {
		// Handles cases where the first character is a token. For instance, "2" is a token, but we want an integer.
		handlePossibleGenericOptimToken(new int[]{PLUS, DOT});
		// This section insures that a plus is not followed by a minus.
		if (LA(1) == PLUS && LA(2) == GENERICOPTIMTOKEN) {
			Token token = LT(2);
			if (token.getText().charAt(0) == '-') {
				throw new RecognitionException("Invalid value for floatNumber rule.", getFilename(), token.getLine(), token.getColumn());
			}
		}		
	}
	: (	(PLUS)?
		( 
	    	minusOrInteger (DOT (unsignedInteger)?)? |
	     	DOT unsignedInteger
	   	) 
	  )
	;

/** 
 * A minus sign, unsigned integer, or a minus sign followed by an integer.
 * The action block checks that the characters are valid.
 */		
minusOrInteger {
		Token token = LT(1);
		String tokenText = token.getText();
		boolean isValidToken = true;
		if (tokenText.charAt(0) == '-'){
			tokenText = tokenText.substring(1);
		}
		if (!tokenText.isEmpty() && !isValidUnsignedInteger(tokenText)){
			throw new RecognitionException("Invalid value for signedInteger rule.", getFilename(), token.getLine(), token.getColumn());
		}	
	}
	: genericOptimToken
	;

signedInteger {
		Token token = LT(1);
		String tokenText = token.getText();
		if (tokenText.charAt(0) == '-'){
			tokenText = tokenText.substring(1);
		}
		if (!isValidUnsignedInteger(tokenText)){
			throw new RecognitionException("Invalid value for signedInteger rule.", getFilename(), token.getLine(), token.getColumn());
		}	
	}
	: genericOptimToken
	;

/** 
 * An unsigned integer.
 * Unsigned integers contain a subset of characters that can be in a GENERICOPTIMTOKEN.
 * The action block checks that the characters are valid.
 */		
unsignedInteger {
		Token token = LT(1);
		if (!isValidUnsignedInteger(token)){
			throw new RecognitionException("Invalid value for unsignedInteger rule.", getFilename(), token.getLine(), token.getColumn());
		}
	}
	: genericOptimToken
	;

/**
 * The date according to regional settings.
 * For import/export in Optim Distributed, this is bad.
 * Dates with one format on one machine will not be read correctly by another.
 * According to my tests, the date will always show with three parts.
 * The things that are changeable are:
 *   - The order of the elements.
 *   - The separator character.
 * Common separator characters are '/', '.', and '-'.  Other separators are possible, but for now, we only use these separators.
 * If the '-' separator is used, only one token is read (hence the genericOptimToken option).  
 */
regionalFormatDate
	: unsignedInteger dateSeparator unsignedInteger dateSeparator unsignedInteger |
	  genericOptimToken
	;

dateSeparator
	: SLASH | DOT
	;

/**
 * Standard format date.
 * The format is always YYYY/MM/DD.
 * Other formats must use different rules.
 */
standardFormatDate
	: unsignedInteger SLASH unsignedInteger SLASH unsignedInteger
	;

/*
 * Choices
 */ 
alwaysPromptNeverChoice
	: "A"
	| "P"
	| "N"
	;
 
alwaysSuccessFailureChoice
    : "A"
    | "S"
    | "F"
    ;

andOrChoice
	: "A"
	| "O"
	;

caseChoice
    : "U"
    | "L"
    | "M"
    | "P"
    ;

dataObjectsBothChoice
	: "D"
	| "O"
	| "B"
	;

defaultKeyScanChoice
	: "D"
	| "K"
	| "S"
	;
	
dormantInitialSelectedChoice
	: "D"
	| "I"
	| "S"
	;
 
moveCompareChoice
	: "M"
	| "C"
	;
	
noneGlobalLocalChoice
	: "N"
	| "G"
	| "L"
	;

noneLocalNamedChoice
	: "N"
	| "L"
	| "F"
	;

oldDefaultNewNotChoice
    : "O"
    | "D"
    | "N"
    | "T"
    ;

trueFalseChoice
	: "T"
	| "F"
	;

typeChoice
	: "2"
	| "C"
	| "O"
	| "S"
	| "I"
	| "G"
	| "M"
	| "T"
	| "Z"
	;

yesNoChoice
    : "Y"
    | "N"
    ;

/*
 * Methods
 */
commitDeleteMethod
    : "E"
    | "C"
    ;

deleteBeforeInsertMethod
    : "N"
    | "A"
    | "M"
    ;

insertProcessMethod
    : "I"
    | "O"
    | "U"
    | "M"
    ;
 
insertTableMethod
    : "Y"
    | "N"
    | "O"
    ;

/*
 * Types
 */
accessStrategyType
    : "D"
    | "K"
    | "S"
    ;
 
ageType
	: "NONE"
	| "INCREMENTAL"
	| "RULEBASED"
	| "SPECIFICYEAR"
	| "SPECIFICDATE"
	| "TARGETDATES"
	;
 
archiveActionType
	: "SEP"
	| "BEFRT"
	| "BER"
	| "AELRT"
	| "EEP"
	| "SDP"
	| "BDFRT"
	| "BDR"
	| "ADR"
	| "ADLRT"
	| "EDP"
	| "SRP"
	| "BRFRT"
	| "BRR"
	| "ARR"
	| "ARLRT"
	| "ERP"
	| "AELT"
	;
 
errorType
	: "PROCESS"
	| "SKIP"
	| "STOP"
	;

headingType
	: "NL"
	| "NC"
	| "NR"
	| "LL"
	| "LC"
	| "LR"
	;

indexTablespaceType
    : "TABLE"
    | "SOURCE"
    | "EXPLICIT"
    ;

indexType
	: "TABLE"
	| "SOURCE"
	| "CURRENT"
	| "EXPLICIT"
	;

dbAliasObjectCreatorIDType
	: "OBJECT"
	| "SOURCE"
	| "CURRENT"
	| "EXPLICIT"
	;

mapSourceType
    : "X"
    | "A"
    | "L"
    | "D"
    ;

objectType
	: "ASSEMBLY"
	| "DEFAULT"
	| "FUNCTION"
	| "PACKAGE"
	| "PARTITION_FUNCTION"
	| "PARTITION_SCHEME"
	| "PROCEDURE"
	| "RULE"
	| "SEQUENCE"
	| "UDT"
	| "VIEW"
	;

processingMode
	: "S" | "P"
	;

restoreActionType returns [boolean isRestoreActionInsert = false]
	: "INSERT" {isRestoreActionInsert = true;}
	| "LOAD"   {isRestoreActionInsert = false;}
	;

requestSelectionMode
	: "DATAMODEL"
	| "GROUP"
	| "DESC"
	| "DATE"
	;

statusType
	: "K"
	| "NEWK"
	| "UNK"
	| "NEWUNK"
	;

/*
 * Operator
 */
operator
    : EQUALS
    | (LESSTHAN (EQUALS | GREATERTHAN)?)
    | (GREATERTHAN (EQUALS)?)
    | (BANG (EQUALS | LESSTHAN | GREATERTHAN))
    | (CARET (EQUALS | LESSTHAN | GREATERTHAN))
    | "EQ"
    | "LT"
    | "GT"
    | "LE"
    | "GE"
    | "NE"
    | "IN"
    | "BETWEEN"
    | "LIKE"
    | QUOTEDSTRING
    ;

/**
 * Variable Delimiter used in Archive Actions and Access Definition tables.
 * NOTE: GENERICOPTIMTOKEN replaces the tokens AT, POUND, and DOLLAR since an GENERICOPTIMTOKEN
 * can be one of these tokens.
 * A GENERICOPTIMTOKEN can also be '0'.  Some rules allow the zero, others do not.  Therefore, the rule is parameterized.
 */
variableDelimiter[boolean allowZero] {
		/*
		 * If the token is an GENERICOPTIMTOKEN, make sure it's a value that is legal for variableDelimiter.
		 */
		if (LA(1) == GENERICOPTIMTOKEN) {
			Token token = LT(1);
			if(!isValidVariableDelimiter(token, allowZero)){
				throw new RecognitionException("Invalid value for variableDelimiter rule.", getFilename(), token.getLine(), token.getColumn());
			}
		}
	}
	: TILDE
	| BANG
	| GENERICOPTIMTOKEN
	| PERCENT
	| AMPERSAND
	| PLUS
	| COLON
	| QUESTION
	;

/**
 * A string delimiter character (either a single quote or double quote) enclosed in parentheses.
 * NOTE: This rule cannot use the tokens SINGLEQUOTE or DOUBLEQUOTE.  Otherwise,
 * the parser will attempt to read a QUOTEDSTRING, and it will fail because there is only
 * one quote character.
 *
 * Instead, we define tokens that represent the entire parenthesized delimiter.
 * See the token definitions for additional information.
 * Readers of the tree must know how to handle these tokens - the tree will contain one single token, 
 * *not* three tokens representing the left parenthesis, the delimiter character, and the right parenthesis. 
 * 
 * Originally, I tried using literals using quoted strings like "\'".  However, this didn't work
 * because the lexer still attempted to read the token as a quoted string.
 * From the ANTLR 2.7.7 online help:
 *
 *    "The literals table (which is usally implemented as a hash table in the lexer) 
 *    is checked after each token is matched, so that the literals effectively override 
 *    the more general identifier pattern."
 *
 * In other words, the tokens are determined *first*, and *then* literals are used to override.
 * This is why the literal "\'" doesn't work - the lexer finds the quoted string first, and never
 * has a chance to override it.
 *
 * Similarly, using a literal with embedded parentheses did not work because we have tokens for the parentheses.
 *
 * By contrast, if we use the literal "+" along with a defined PLUS token, the literal would override the PLUS token
 * because the lexer recognizes the single token.
 *
 * The only solution to the problem was to define tokens that represented exactly what we wanted.
 *
 * There is a potential hazard - if there is a quoted string that appears inside parentheses like this:
 *
 *    ("this is a quoted string")
 *
 * Then there is the possibility for error if the quoted string begins with a right parenthesis:
 *
 *    (") is a parenthesis character")
 *
 * Likelihood of this scenario: very low.
 */
parenthesizedStringDelimiter 
	: 
	PARENTHESIZEDSINGLEQUOTE | PARENTHESIZEDDOUBLEQUOTE 
	;

/**
 * Decimal separator
 */
decimalSeparator
	: DOT
	| COMMA
	;

/*
 * OEF Distributed Lexer
 */
{ @SuppressWarnings("all") }
class DistributedLexer extends Lexer;

options {
	k = 5;
	charVocabulary = '\u0000'..'\uFFFE';
}

/**
  NOTE: The QUOTEDSTRING token is added because the last line:
  
  	 ~('\n'|'\r')
  	 
  Handles individual characters, including quote characters. Quoted strings need to be handled
  as quoted strings because the string might contain a literal "//" that is not meant to
  terminate the delimited string.
*/
DELIMITEDSTRING
	: "//"
	(
		options {
			greedy = false;  // Make it exit upon "//"
			generateAmbigWarnings = false; // Shut off newline errors
		}
		:	'\r' '\n'	{newline();}
		|	'\r'		{newline();}
		|	'\n'		{newline();}
		|	QUOTEDSTRING
		|	~('\n'|'\r')
		)*
	"//"
	;
 
/**
 * Used for identifiers such as definition IDs and names.
 * A string of one or more characters where special characters are excluded.
 * The order of the characters after the '%' comes from rtrdba.cpp gInvalidChars.
 * This token is also used for integers.
 *
 * NOTE: We cannot use a separate token for integers.
 */
GENERICOPTIMTOKEN
	: (
		~(
			' ' | 
			'\t'| 
			'\u001a' |
			'\n' |
			'\r' |
			'%' |
			'~' |
			'`' |
			'!' |
			'^' |
			'&' |
			'*' |
			'(' |
			')' |
			'+' |
			'=' |
			'|' |
			'{' |
			'}' |
			'[' |
			']' |
			':' |
			';' |
			'\"' |
			'\'' |
			'<' |
			'>' |
			',' |
			'.' |
			'?' |
			'/' |
			'\\'			
	 	)
	)+
	;

/**
 * A parenthesized single quote.
 * This must have higher precendence than LEFTPAREN alone.
 */
PARENTHESIZEDSINGLEQUOTE
	: LEFTPAREN SINGLEQUOTE RIGHTPAREN
	;

/**
 * A bogus copy of a token.
 * This is a hack to make ANTLR handle our grammar nicely.
 * Without this token defined, ANTLR only needs to look ahead two characters
 * to disambiguate PARENTHESIZEDSINGLEQUOTE, PARENTHESIZEDDOUBLEQUOTE, and LEFTPAREN.
 * However, if ANTLR only looks ahead two characters, then it misreads this:
 *
 *       ('A')
 * 
 * and treats "('A" as a PARENTHESIZEDSINGLEQUOTE instead of reading this as 
 * LEFTPAREN, QUOTEDSTRING, RIGHTPAREN.
 *
 * This token causes ANTLR to generate code to look ahead three characters, properly 
 * disambiguating the above sequence.  
 * Even though this token is identical to PARENTHESIZEDSINGLEQUOTE, it will never
 * be used because it is defined after PARENTHESIZEDSINGLEQUOTE.  
 * PARENTHESIZEDSINGLEQUOTE takes precedence.
 */
BOGUSPARENTHESIZEDSINGLEQUOTE
	: PARENTHESIZEDSINGLEQUOTE
	;

/**
 * A parenthesized double quote.
 * This must have higher precendence than LEFTPAREN alone.
 */
PARENTHESIZEDDOUBLEQUOTE
	: LEFTPAREN DOUBLEQUOTE RIGHTPAREN
	;

/**
 * A bogus copy of a token.
 * See BOGUSPARENTHESIZEDSINGLEQUOTE for a complete explanation.
 */
BOGUSPARENTHESIZEDDOUBLEQUOTE
	: PARENTHESIZEDDOUBLEQUOTE
	;

QUOTEDSTRING
	: (((DOUBLEQUOTE)(~'"')*(DOUBLEQUOTE)) | ((SINGLEQUOTE)(~'\'')*(SINGLEQUOTE)))
	;

AMPERSAND
	: '&'
	;

PIPE
	: '|'
	;

BACKSLASH
	: '\\'
	;

BANG
	: '!'
	;

CARET
	: '^'
	;

COLON
	: ':'
	;

COMMA
	: ','
	;

DOT
	: '.'
	;

EQUALS
	: '='
	;

GREATERTHAN
	: '>'
	;

LEFTPAREN
	: '('
	;

/*
LEFTPARENFAKEOUT
	: '('.')'
	;
*/

LESSTHAN
	: '<' 
	;

PERCENT
	: '%'
	;

PLUS
	: '+'
	;

QUESTION
	: '?'
	;

RIGHTPAREN
	: ')'
	;

SEMI
	: ';'
	;

STAR
	: '*'
	;

protected
DOUBLEQUOTE
	: '"'
	;

protected
SINGLEQUOTE
	: '\''
	;

TILDE
	: '~'
	;

SLASH
	: '/'
	;

NEWLINE
	: ("\r\n" | '\n' | '\r') {$setType(Token.SKIP); newline();}
	;
WS
	: (' ' | '\t' | '\u001a') {$setType(Token.SKIP);}
	;

